import logging
import os
import tempfile
import unittest

import matplotlib.pyplot as plt
import pandas as pd
from funman_demo.handlers import RealtimeResultPlotter, ResultCacheWriter
from pysmt.shortcuts import (
    FALSE,
    GE,
    GT,
    LE,
    LT,
    TRUE,
    And,
    Equals,
    ForAll,
    Function,
    FunctionType,
    Iff,
    Int,
    Plus,
    Real,
    Symbol,
    get_model,
    simplify,
    substitute,
)
from pysmt.typing import BOOL, INT, REAL

from funman import Funman
from funman.model import Model, Parameter, QueryLE
from funman.representation.representation import (
    Point,
    ResultCombinedHandler,
    SearchConfig,
)
from funman.scenario.consistency import ConsistencyScenario
from funman.scenario.parameter_synthesis import ParameterSynthesisScenario
from funman.search import BoxSearch, SMTCheck
from funman.translate.bilayer import (
    Bilayer,
    BilayerEncoder,
    BilayerEncodingOptions,
    BilayerMeasurement,
    BilayerModel,
)
from funman.utils.smtlib_utils import smtlibscript_from_formula

l = logging.getLogger(__file__)
l.setLevel(logging.ERROR)

DATA = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), "../../../resources/bilayer"
)


class TestChimeBilayerSolve(unittest.TestCase):
    def setup(self, duration=10, transmission_reduction=0.05):
        bilayer_json_file = os.path.join(
            DATA, "CHIME_SIR_dynamics_BiLayer.json"
        )
        bilayer = Bilayer.from_json(bilayer_json_file)
        assert bilayer

        measurements = {
            "state": [{"variable": "I"}],
            "observable": [{"observable": "H"}],
            "rate": [{"parameter": "hr"}],
            "Din": [{"variable": 1, "parameter": 1}],
            "Dout": [{"parameter": 1, "observable": 1}],
        }
        hospital_measurements = BilayerMeasurement.from_json(measurements)

        model = BilayerModel(
            bilayer,
            measurements=hospital_measurements,
            init_values={"S": 10000, "I": 1, "R": 1},
            parameter_bounds={
                "beta": [
                    0.000067,
                    0.000067,
                ],
                # "beta" : [0.00005, 0.00007],
                "gamma": [1.0 / 14.0, 1.0 / 14.0],
                "hr": [0.01, 0.01],
            },
        )

        if isinstance(transmission_reduction, list):
            lb = model.parameter_bounds["beta"][0] * (
                1.0 - transmission_reduction[1]
            )
            ub = model.parameter_bounds["beta"][1] * (
                1.0 - transmission_reduction[0]
            )
        else:
            lb = model.parameter_bounds["beta"][0] * (
                1.0 - transmission_reduction
            )
            ub = model.parameter_bounds["beta"][1] * (
                1.0 - transmission_reduction
            )
        model.parameter_bounds["beta"] = [lb, ub]

        query = QueryLE("H", 0.5)

        encoder = BilayerEncoder(
            config=BilayerEncodingOptions(step_size=1, max_steps=duration)
        )

        return model, query, encoder

    # @unittest.skip("temporarily remove")
    def test_chime_bilayer_solve(self):
        model, query, encoder = self.setup(
            duration=1, transmission_reduction=0.00
        )

        query.ub = 10000

        scenario = ConsistencyScenario(model, query, _smt_encoder=encoder)

        result = Funman().solve(
            scenario, config=SearchConfig(solver="dreal", search=SMTCheck)
        )
        assert result

        result.plot(logy=True)
        print(result.dataframe())

    # @unittest.skip("temporarily remove")
    def test_chime_bilayer_synthesize(self):
        model, query, encoder = self.setup(
            duration=10, transmission_reduction=[-0.05, 0.1]
        )
        model.init_values = {
            "S": 7438.567991,
            "I": 2261.927694,
            "R": 299.504315,
        }
        query.ub = 75
        # The efficacy can be up to 4x that of baseline (i.e., 0.05 - 0.20)
        parameters = [
            Parameter(
                name="beta",
                # lb=0.000001,
                # ub=0.00001,
                lb=model.parameter_bounds["beta"][0],
                ub=model.parameter_bounds["beta"][1],
            )
        ]
        tmp_dir_path = tempfile.mkdtemp(prefix="funman-")
        result = Funman().solve(
            ParameterSynthesisScenario(
                parameters, model, query, _smt_encoder=encoder
            ),
            config=SearchConfig(
                number_of_processes=1,
                tolerance=1e-8,
                solver="dreal",
                search=BoxSearch,
                handler=ResultCombinedHandler(
                    [
                        ResultCacheWriter(
                            os.path.join(tmp_dir_path, "search.json")
                        ),
                        RealtimeResultPlotter(
                            parameters,
                            plot_points=True,
                            realtime_save_path=os.path.join(
                                tmp_dir_path, "search.png"
                            ),
                        ),
                    ]
                ),
            ),
        )
        assert result

        # sample points from true boxes and call
        ps = result.parameter_space
        points = []
        for tbox in ps.true_boxes:
            values = {}
            for p, i in tbox.bounds.items():
                param_assignment = (i.lb + i.ub) * 0.5
                values[p.name] = param_assignment
            points.append(Point.from_dict({"values": values}))

        dfs = result.true_point_timeseries(points=points)
        for point, df in zip(points, dfs):
            print("-" * 80)
            print("Parameter assignments:")
            for param, value in point.values.items():
                print(f"    {param.name} = {value}")
            print("-" * 80)
            print(df)
            print("=" * 80)


if __name__ == "__main__":
    unittest.main()



import io
import os
from functools import partial
from queue import Queue
from typing import Dict, List

import docker
import dreal
import pysmt.smtlib.commands as smtcmd
from funman_dreal.converter import DRealConverter
from pysmt.decorators import clear_pending_pop
from pysmt.exceptions import (
    SolverRedefinitionError,
    SolverReturnedUnknownResultError,
    UnknownSolverAnswerError,
)
from pysmt.logics import QF_NRA
from pysmt.shortcuts import Real, get_env
from pysmt.smtlib.parser import SmtLibParser
from pysmt.smtlib.script import SmtLibCommand
from pysmt.smtlib.solver import SmtLibOptions, SmtLibSolver
from pysmt.solvers.eager import EagerModel
from pysmt.solvers.smtlib import SmtLibBasicSolver, SmtLibIgnoreMixin
from pysmt.solvers.solver import Solver, SolverOptions
from tenacity import retry

from funman.utils.smtlib_utils import FUNMANSmtPrinter

# def setup(smt2_file, benchmark_path, out_dir):
#     print("setting up docker")
#     if os.path.exists(out_dir):
#         rmtree(out_dir)
#     os.makedirs(out_dir)
#     copyfile(
#         os.path.join(benchmark_path, smt2_file),
#         os.path.join(out_dir, smt2_file),
#     )


# TODO find a better way to determine if solver was successful
def run_dreal(
    smt2_file,
    # benchmark_path=os.path.join(os.path.abspath("../../resources/smt2")),
    out_dir="out",
    solver_opts=" --precision 1 --verbose debug",
):
    # setup(smt2_file, benchmark_path, out_dir)

    smt2_file_path = os.path.dirname(smt2_file)
    smt2_filename = os.path.basename(smt2_file)

    volumes = {
        smt2_file_path: {"bind": "/smt2", "mode": "rw"},
        out_dir: {"bind": "/out", "mode": "rw"},
        "/var/run/docker.sock": {"bind": "/var/run/docker.sock", "mode": "rw"},
    }
    client = docker.from_env()
    container = client.containers.run(
        "dreal/dreal4",
        command=f"dreal /smt2/{smt2_filename} {solver_opts}",
        # command="/bin/bash",
        volumes=volumes,
        detach=True,
    )
    container.wait()
    result = container.logs()

    with open(os.path.join(out_dir, f"{smt2_filename}.stdout"), "w") as f:
        f.write(result.decode())
        f.close()

    if not os.path.exists(os.path.join(out_dir, f"{smt2_filename}.model")):
        container.stop()
        ll_client = docker.APIClient()
        ll_client.remove_container(container.id)
        raise Exception("Could not construct a model")

    container.stop()
    ll_client = docker.APIClient()
    ll_client.remove_container(container.id)

    with open(os.path.join(out_dir, f"{smt2_filename}.model")) as f:
        model = f.read()

    return model


def add_dreal_to_pysmt():
    name = "dreal"
    env = get_env()
    if name in env.factory._all_solvers:
        raise SolverRedefinitionError("Solver %s already defined" % name)
    solver = partial(DRealNative, "dreal/dreal4", LOGICS=DRealNative.LOGICS)
    solver.LOGICS = DRealNative.LOGICS
    env.factory._all_solvers[name] = solver
    env.factory._generic_solvers[name] = ("dreal/dreal4", DRealNative.LOGICS)
    env.factory.solver_preference_list.append(name)


def ensure_dreal_in_pysmt():
    try:
        add_dreal_to_pysmt()
    except SolverRedefinitionError:
        pass


class DReal(SmtLibSolver):
    LOGICS = [QF_NRA]
    OptionsClass = SmtLibOptions

    commands_to_enqueue = [
        smtcmd.SET_LOGIC,
        smtcmd.PUSH,
        smtcmd.POP,
        smtcmd.DEFINE_FUN,
        smtcmd.ASSERT,
    ]
    commands_to_flush = [
        smtcmd.CHECK_SAT,
    ]

    def __init__(
        self, args, environment, logic, LOGICS=None, **options
    ) -> None:
        Solver.__init__(self, environment, logic=logic, **options)
        self.to = self.environment.typeso
        self.LOGICS = DReal.LOGICS
        self.declared_vars = [set()]
        self.declared_sorts = [set()]
        self.parser = SmtLibParser(interactive=True)

        # solver_opts = kwargs["solver_opts"] if "solver_opts" in kwargs else ""
        self.batch = Queue()  # Batch of commands to send in one call
        self.sent_batch = (
            Queue()
        )  # Batch of commands already sent and needs to be checked
        self.wait = 1  # How long to wait on next socket recv
        self.last_sent = None  # Last message sent to container
        self.client = docker.APIClient()
        self.container = self.client.create_container(
            "dreal/dreal4",
            stdin_open=True,
            tty=True,
            entrypoint=f"/usr/bin/dreal --in --model --smtlib2-compliant",
        )
        self.client.start(self.container)

        self.stdout = self.client.attach_socket(
            self.container,
            params={"stdout": 1, "stderr": 1, "stream": 1},
        )
        self.stdin = self.client.attach_socket(
            self.container,
            params={"stdin": 1, "stream": 1},
        )

        self.models: List[Dict[str, (float, float)]] = []
        self.current_model = None
        self.symbols = set([])

        # Initialize solver
        try:
            self.options(self)
        except Exception as e:
            pass

        self.set_logic(logic)

    def container_alive(func):
        def inner(*args):
            # time.sleep(1)
            status = args[0].client.inspect_container(args[0].container)[
                "State"
            ]["Status"]
            if status == "exited":
                raise Exception(
                    f"Cannot send command to container with status: {status}"
                )
            else:
                return func(*args)

        return inner

    # @container_alive
    def _send_command(self, cmd):
        """Sends a command to the socket."""
        self._debug("Sending: %s", cmd.serialize_to_string())

        # s._sock.send(input.encode("utf-8"))

        if cmd.name == smtcmd.SET_OPTION:
            self.wait = 0
            return
        else:
            self.wait = 1

            if cmd.name == smtcmd.CHECK_SAT:
                self.current_model = {}
                self.models.append(self.current_model)
            elif cmd.name == smtcmd.DECLARE_FUN:
                symbol = cmd.args[0]
                self.symbols.add(symbol)

            # self.last_sent = payload
            self.batch.put(cmd)
            if cmd.name in DReal.commands_to_flush:
                self._send_batch()

            # print(f"Sent: {payload}")

    def _get_value_answer(self):
        """Reads and parses an assignment from the STDOUT pipe"""
        value = self.solver_stdout
        lst = self.parser.get_assignment_list(value)
        # lst = [[0.0, 0.0]]
        self._debug("Read: %s", lst)
        return lst

    def get_value(self, item):
        return Real(
            self.current_model[item.symbol_name()][0]
        )  # return lower bound

    def _send_batch(self):
        payload = b""
        while not self.batch.empty():
            cmd = self.batch.get()
            text = io.StringIO()
            cmd.serialize(text, daggify=False, printer=FUNMANSmtPrinter(text))
            text.write("\r\n")
            encoded_text = text.getvalue().encode("utf-8")
            payload += encoded_text
            self.sent_batch.put((cmd, encoded_text))
        # payload += b"\r\n"
        self.stdin._sock.send(payload)
        # self.stdin.flush()

    @retry
    def _read_socket(self):
        return self.stdout._sock.recv(1)

    def _read_line(self):
        # Read a line of content, skipping empty lines with only "\n"
        msg = []
        while True:
            b = self._read_socket()
            if b == b"\r" or (b == b"\n" and len(msg) == 0):
                continue
            msg.append(b)
            if b == b"\n":
                line = b"".join(msg)
                # print(f"Read: {line}")
                return line

    def _get_echo_result(self, encoded_text):
        """
        Read back the stdout from dreal that echo's the encoded_text

        Parameters
        ----------
        command : str
            SMTLib Command
        encoded_text : str
            Serialized command text
        """

        encoded_text = encoded_text.replace(b"\r", b"")
        line = self._read_line()  # .replace(b"\n", b"")
        if encoded_text == line:
            return "success"
        else:
            raise Exception(
                f"Could not read back echoed command from dreal: {encoded_text}, Got: {line}"
            )

    def get_check_sat_value_result(self):
        line = self._read_line().decode()
        if "delta-sat" in line:
            return "delta-sat"
        elif "unsat" in line:
            return "unsat"
        elif "sat" in line:
            return "sat"

    def _get_model(self):
        """
        Read list of "symbol: [lb, ub]\n" and store as model
        """
        while len(self.current_model) < len(self.symbols):
            line = self._read_line().decode()
            [symbol_str, bounds] = line.split(":")
            [lb, ub] = (
                bounds.replace("[", "").replace("]", "").strip().split(",")
            )
            self.current_model[symbol_str.strip()] = [float(lb), float(ub)]

    def _get_check_sat_result(self, encoded_text):
        """
        Read back the stdout from dreal that echo's the encoded_text, the result, and possibly a model.

        Parameters
        ----------
        command : str
            SMTLib Command
        encoded_text : str
            Serialized command text
        """
        if self._get_echo_result(encoded_text):  # read back "(check-sat)"
            result = self.get_check_sat_value_result()
            if result == "delta-sat" or result == "sat":
                self._get_model()
            return result

    def _get_command_result(self, command, encoded_text):
        handlers = {
            smtcmd.SET_LOGIC: self._get_echo_result,
            smtcmd.DECLARE_FUN: self._get_echo_result,
            smtcmd.ASSERT: self._get_echo_result,
            smtcmd.CHECK_SAT: self._get_check_sat_result,
        }
        try:
            return handlers[command.name](encoded_text)
        except TimeoutError as e:
            print(f"Timeout waiting for socket receive.")
            res = "success"
        except Exception as e:
            print(f"Failed to read from socket: {e}")

    # @container_alive
    def _get_answer(self):
        """Reads a line from socket."""
        if self.wait > 0:
            # print("Reading from socket: ")
            # self.stdout._sock.settimeout(self.wait)
            res = "success"
            while not self.sent_batch.empty():
                res = self._get_command_result(*self.sent_batch.get())
        else:
            res = "success"
        # print(f"Read: {res}")
        return res

    @clear_pending_pop
    def solve(self, assumptions=None):
        assert assumptions is None
        self._send_command(SmtLibCommand(smtcmd.CHECK_SAT, []))
        ans = self._get_answer()
        if ans == "sat" or ans == "success" or "delta-sat" in ans:
            return True
        elif "unsat" in ans:
            return False
        elif ans == "unknown":
            raise SolverReturnedUnknownResultError
        else:
            raise UnknownSolverAnswerError("Solver returned: " + ans)

    def _exit(self):
        # self._send_command(SmtLibCommand(smtcmd.EXIT, []))

        self.stdin.close()
        self.stdout.close()
        print("Stopping dreal container ...")
        self.client.stop(self.container)
        self.client.wait(self.container)
        # raw_stream,status = client.get_archive(container,'/received.txt')
        # tar_archive = BytesIO(b"".join((i for i in raw_stream)))
        # t = tarfile.open(mode='r:', fileobj=tar_archive)
        # text_from_container_file = t.extractfile('received.txt').read().decode('utf-8')
        self.client.remove_container(self.container)
        print("Done stopping dreal container.")
        return

    def send_input(self, input):
        s = self.client.attach_socket(
            self.container,
            params={"stdout": 1, "stderr": 1, "stream": 1, "stdin": 1},
        )
        # pass
        s._sock.settimeout(1)
        s._sock.send(input.encode("utf-8"))
        rval = io.StringIO()
        while 1:
            try:
                msg = s._sock.recv(1024)
            except TimeoutError as e:
                # print(f"Timeout waiting for data from container: {e}")
                msg = None
            if not msg:
                break
            else:
                try:
                    for line in msg.split(b"\n"):
                        content = f"{line[8:].decode()}\n"
                        rval.write(content)
                except Exception as e:
                    print(f"Could not decode response: {msg} because {e}")
        s.close()
        return rval.getvalue()


class DRealNative(Solver, SmtLibBasicSolver, SmtLibIgnoreMixin):
    LOGICS = [QF_NRA]
    OptionsClass = SolverOptions

    commands_to_enqueue = [
        smtcmd.SET_LOGIC,
        smtcmd.DEFINE_FUN,
        smtcmd.PUSH,
        smtcmd.POP,
    ]
    commands_to_flush = [smtcmd.CHECK_SAT, smtcmd.ASSERT]

    def __init__(
        self, args, environment, logic, LOGICS=None, **options
    ) -> None:
        Solver.__init__(self, environment, logic=logic, **options)

        # Setup super class attributes
        self.converter = DRealConverter(environment)
        # self.options(self)
        # self.options.set_params(self)
        self.mgr = environment.formula_manager
        self.model = None

        # dreal specific attributes
        self.config = dreal.Config()

        if "dreal_precision" in options:
            self.config.precision = options["solver_options"][
                "dreal_precision"
            ]
        # self.context.config.use_polytope = True
        # self.context.config.use_worklist_fixpoint = True
        self.model = None
        if "dreal_log_level" in options["solver_options"]:
            if options["solver_options"]["dreal_log_level"] == "debug":
                dreal.set_log_level(dreal.LogLevel.DEBUG)
            elif options["solver_options"]["dreal_log_level"] == "info":
                dreal.set_log_level(dreal.LogLevel.INFO)

        if (
            "dreal_mcts" in options["solver_options"]
            and options["solver_options"]["dreal_mcts"]
        ):
            self.config.mcts = True

        self.context = dreal.Context(self.config)
        self.context.SetLogic(dreal.Logic.QF_NRA)

        # self.to = self.environment.typeso
        self.LOGICS = DReal.LOGICS
        self.symbols = {}

    def __del__(self):
        self.context.Exit()

    @clear_pending_pop
    def add_assertion(self, formula, named=None):
        # print(f"Assert({formula})")

        f = self.converter.convert(formula)

        deps = formula.get_free_variables()
        # Declare all variables
        for symbol in deps:
            assert symbol.is_symbol()
            self.cmd_declare_fun(
                SmtLibCommand(name=smtcmd.DECLARE_FUN, args=[symbol])
            )
        code = self.context.Assert(f)

    def cmd_set_option(self, cmd):
        pass

    def cmd_set_logic(self, cmd):
        pass

    def cmd_declare_fun(self, cmd):
        # print(f"DeclareVariable({cmd.args[0]})")
        if cmd.args[0] in self.converter.symbol_to_decl:
            v = self.converter.symbol_to_decl[cmd.args[0]]
        else:
            v = dreal.Variable(cmd.args[0].symbol_name(), dreal.Variable.Real)
        self.context.DeclareVariable(v)
        self.symbols[cmd.args[0].symbol_name()] = (cmd.args[0], v)

    def cmd_assert(self, cmd: SmtLibCommand):
        self.add_assertion(cmd.args[0])

    def cmd_push(self, cmd):
        self.push(cmd.args[0])

    def push(self, levels):
        # print("Push()")
        self.context.Push(levels)

    def cmd_pop(self, cmd):
        self.pop(cmd.args[0])

    def pop(self, levels):
        # print("Pop()")
        self.context.Pop(levels)

    def cmd_check_sat(self, cmd):
        return self.check_sat()

    def check_sat(self):
        # print("CheckSat()")
        result = self.context.CheckSat()
        # result = dreal.CheckSatisfiability(self.assertion, 0.001)
        self.model = result
        return result

    def _send_command(self, cmd):
        handlers = {
            smtcmd.SET_LOGIC: self.cmd_set_logic,
            smtcmd.SET_OPTION: self.cmd_set_option,
            smtcmd.DECLARE_FUN: self.cmd_declare_fun,
            smtcmd.ASSERT: self.cmd_assert,
            smtcmd.PUSH: self.cmd_push,
            smtcmd.POP: self.cmd_pop,
            smtcmd.CHECK_SAT: self.cmd_check_sat,
        }
        return handlers[cmd.name](cmd)

    def get_model(self):
        assignment = {}
        for sn in self.symbols:
            s = self.symbols[sn][0]
            if s.is_term():
                v = self.get_value(self.symbols[sn][1])
                assignment[s] = v
        return EagerModel(assignment=assignment, environment=self.environment)

    def get_value(self, item):
        # print(f"get_value() {item}: {self.model[item]}")
        return Real(self.model[item].lb())

    @clear_pending_pop
    def solve(self, assumptions=None):
        assert assumptions is None
        try:
            ans = self.check_sat()
        except Exception as e:
            raise e
        return ans is not None

    def _exit(self):
        pass



import unittest

import funman_dreal
from pysmt.logics import QF_NRA
from pysmt.shortcuts import (
    LE,
    LT,
    REAL,
    And,
    Equals,
    ForAll,
    Not,
    Or,
    Plus,
    Pow,
    Real,
    Solver,
    Symbol,
)

# from pysmt.smtlib.script import smtlibscript_from_formula


class TestRunDrealNative(unittest.TestCase):
    def test_dreal(self):
        # Needed to use dreal with pysmt
        funman_dreal.ensure_dreal_in_pysmt()

        # x = Variable("x")
        # y = Variable("y")
        # z = Variable("z")

        # f_sat = And(0 <= x, x <= 10, 0 <= y, y <= 10, 0 <= z, z <= 10, sin(x) + cos(y) == z)

        # result = CheckSatisfiability(f_sat, 0.001)
        # print(result)

        with Solver(name="dreal", logic=QF_NRA) as s:
            x = Symbol("x", REAL)
            y = Symbol("y", REAL)
            z = Symbol("z", REAL)

            f_sat = And(
                [
                    LE(Real(0.0), x),
                    LE(x, Real(10.0)),
                    LE(Real(0.0), y),
                    LE(y, Real(10.0)),
                    LE(Real(0.0), z),
                    LE(z, Real(10.0)),
                    Equals(
                        Plus(Pow(x, Real(2.0)), Pow(y, Real(2.0))),
                        # Plus(x, y),
                        # Real(1.0)
                        Pow(z, Real(2.0)),
                    ),
                ]
            )
            s.add_assertion(f_sat)

            result = None
            if s.solve():
                result = s.get_model()
            print(f"Result: {result}")

            s.push(1)

            a = Symbol("a", REAL)
            g_sat = And(Equals(Plus(a, z), Real(1.0)))

            s.add_assertion(g_sat)
            result = None
            if s.solve():
                result = s.get_model()
            print(f"Result: {result}")

            s.pop(1)

            q_sat = Not(Equals(Plus(a, z), Real(1.0)))
            s.add_assertion(q_sat)
            result = None
            if s.solve():
                result = s.get_model()

            print(f"Result: {result}")

    def test_ea_dreal(self):
        # Needed to use dreal with pysmt
        funman_dreal.ensure_dreal_in_pysmt()

        with Solver(name="dreal", logic=QF_NRA) as s:
            x = Symbol("x", REAL)
            x_lb = Symbol("x_lb", REAL)
            x_ub = Symbol("x_ub", REAL)
            y = Symbol("y", REAL)
            z = Symbol("z", REAL)

            ea = ForAll(
                [x],
                And(
                    [
                        And(LE(Real(0.0), x_ub), LT(x_ub, Real(5.0))),
                        And(LE(Real(0.0), x_lb), LT(x_lb, Real(5.0))),
                        LT(x_lb, x_ub),
                        Or(
                            Not(And(LT(x, x_ub), LE(x_lb, x))),
                            Equals(x, Plus(y, z)),
                        ),
                    ]
                ),
            )
            s.push(1)
            s.add_assertion(ea)

            result = None
            if s.solve():
                result = s.get_model()
                print("Model:")
                for v in ea.get_free_variables():
                    print(f"{v} = {float(result.get_py_value(v))}")
                    if v == x_lb:
                        old_x_lb = result.get_py_value(v)
                    if v == x_ub:
                        old_x_ub = result.get_py_value(v)
            else:
                print("Unsat")

            # s.pop(1)

            # ea1 = ForAll(
            #     [x],
            #     And(
            #         [
            #             # convert interval [x_lb, x_ub] from previous call to no-good
            #             Not(
            #                 And(
            #                     LE(Real(1.2495), x_ub),
            #                     LE(x_ub, Real(3.7495000000000003)),
            #                 )
            #             ),
            #             Not(
            #                 And(
            #                     LE(Real(1.2495), x_lb),
            #                     LE(x_lb, Real(3.7495000000000003)),
            #                 )
            #             ),
            #             And(LE(Real(0.0), x_ub), LE(x_ub, Real(5.0))),
            #             And(LE(Real(0.0), x_lb), LE(x_lb, Real(5.0))),
            #             Or(
            #                 Not(And(LE(x, x_ub), LE(x_lb, x))),
            #                 Equals(x, Plus(y, z)),
            #             ),
            #         ]
            #     ),
            # )
            ea1 = And(
                [
                    # convert interval [x_lb, x_ub] from previous call to no-good
                    Not(
                        And(
                            LE(Real(old_x_lb), x_ub),
                            LT(x_ub, Real(old_x_ub)),
                        )
                    ),
                    Not(
                        And(
                            LE(Real(old_x_lb), x_lb),
                            LT(x_lb, Real(old_x_ub)),
                        )
                    ),
                ]
            )

            s.push(1)
            s.add_assertion(ea1)

            result = None
            if s.solve():
                result = s.get_model()
                print("Model:")
                for v in ea1.get_free_variables():
                    print(f"{v} = {float(result.get_py_value(v))}")
            else:
                print("Unsat")
            # print(f"Result: {result}")


if __name__ == "__main__":
    unittest.main()



import os
import unittest

from funman_demo.example.chime import CHIME
from funman_dreal.solver import run_dreal

from funman.utils.smtlib_utils import (
    smtlibscript_from_formula,
    smtlibscript_from_formula_list,
)

# from pysmt.smtlib.script import smtlibscript_from_formula


class TestRunDreal(unittest.TestCase):
    def get_flat_encoding(self, num_timepoints):
        chime = CHIME()
        vars, (parameters, init, dynamics, query) = chime.make_model(
            assign_betas=True
        )
        phi = chime.encode_time_horizon(
            parameters, init, dynamics, query, num_timepoints
        )
        return phi

    def get_layered_encoding(self, num_timepoints):
        chime = CHIME()
        vars, (parameters, init, dynamics, query) = chime.make_model(
            assign_betas=True
        )
        phi = chime.encode_time_horizon_layered(
            parameters, init, dynamics, query, num_timepoints
        )
        return phi

    def test_dreal(self):
        out_dir = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "out"
        )
        print(out_dir)
        if not os.path.exists(out_dir):
            os.mkdir(out_dir)

        num_timepoints = 2

        phi = self.get_flat_encoding(num_timepoints)
        phis = self.get_layered_encoding(num_timepoints)

        smt2_flat_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            f"../resources/smt2/chime_flat_{num_timepoints}.smt2",
        )
        smt2_layered_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            f"../resources/smt2/chime_layered_{num_timepoints}.smt2",
        )

        with open(smt2_flat_file, "w") as f:
            smtlibscript_from_formula(phi).serialize(f, daggify=False)
        with open(smt2_layered_file, "w") as f:
            smtlibscript_from_formula_list(phis).serialize(f, daggify=False)

        # print(smt2_file)
        result = run_dreal(
            smt2_flat_file,
            out_dir=out_dir,
            solver_opts="",
        )
        # docker run --rm -it -v `pwd`/code/funman/auxiliary_packages/funman_dreal/resources/smt2:/smt2  dreal/dreal4
        # assert result


if __name__ == "__main__":
    unittest.main()



import io
import unittest
from functools import partial

from funman_demo.example.chime import CHIME
from funman_dreal.solver import DReal, run_dreal
from pysmt.exceptions import SolverRedefinitionError
from pysmt.logics import QF_NRA
from pysmt.shortcuts import GT, Equals, Real, Solver, Symbol, get_env
from pysmt.typing import REAL

from funman.utils.smtlib_utils import (
    smtlibscript_from_formula,
    smtlibscript_from_formula_list,
)

# from pysmt.smtlib.script import smtlibscript_from_formula


class TestRunDreal(unittest.TestCase):
    def get_flat_encoding(self, num_timepoints):
        chime = CHIME()
        vars, (parameters, init, dynamics, query) = chime.make_model(
            assign_betas=False
        )
        phi = chime.encode_time_horizon(
            parameters, init, dynamics, query, num_timepoints
        )
        return phi

    def get_layered_encoding(self, num_timepoints):
        chime = CHIME()
        vars, (parameters, init, dynamics, query) = chime.make_model(
            assign_betas=True
        )
        phi = chime.encode_time_horizon_layered(
            parameters, init, dynamics, query, num_timepoints
        )
        return phi

    @unittest.skip
    def test_dreal(self):
        """
        This test runs dreal in docker while iteratively sending commands and keeping the docker container and dreal process alive.
        """
        # out_dir = os.path.join(
        #     os.path.dirname(os.path.abspath(__file__)), "out"
        # )
        # print(out_dir)
        # if not os.path.exists(out_dir):
        #     os.mkdir(out_dir)

        num_timepoints = 2

        phi = self.get_flat_encoding(num_timepoints)
        # phis = self.get_layered_encoding(num_timepoints)

        f = io.StringIO()
        # smtlibscript_from_formula_list(phis).serialize(f, daggify=False)
        smtlibscript_from_formula(phi).serialize(f, daggify=False)

        dreal = DReal()
        # for command in f.getvalue().split("\n"):
        #     print(f"sending {command}")
        #     output = dreal.send_input(command)
        #     print(f"received {output}")

        rval = dreal.send_input(f.getvalue())
        print(f"Received: \n{rval}")

        rval = dreal.send_input(
            """
        (push 1)
        (assert (and (<= i_0 100) (<= i_2 100) ))
        (check-sat)
        """
        )
        print(f"Received: \n{rval}")

        rval = dreal.send_input(
            """
        (push 1)
        (assert (and (<= beta_0 0.00007) (<= 0.00005 beta_0) ))
        (check-sat)
        """
        )
        print(f"Received: \n{rval}")

        rval = dreal.send_input(
            """
        (pop 1)
        (push 1)
        (assert (and (<= beta_0 0.00006) (<= 0.00005 beta_0) ))
        (check-sat)
        """
        )
        print(f"Received: \n{rval}")

        # smt2_flat_file = os.path.join(
        #     os.path.dirname(os.path.abspath(__file__)),
        #     f"../resources/smt2/chime_flat_{num_timepoints}.smt2",
        # )
        # smt2_layered_file = os.path.join(
        #     os.path.dirname(os.path.abspath(__file__)),
        #     f"../resources/smt2/chime_layered_{num_timepoints}.smt2",
        # )

        # with open(smt2_flat_file, "w") as f:
        #     smtlibscript_from_formula(phi).serialize(f, daggify=False)

        # print(smt2_file)
        # result = run_dreal(
        #     smt2_flat_file,
        #     out_dir=out_dir,
        #     solver_opts="",
        # )
        # docker run --rm -it -v `pwd`/code/funman/auxiliary_packages/funman_dreal/resources/smt2:/smt2  dreal/dreal4
        # assert result
        assert rval

    def test_pysmt_to_dreal(self):
        name = "dreal"
        # path = "dreal/dreal4"

        env = get_env()
        if name in env.factory._all_solvers:
            raise SolverRedefinitionError("Solver %s already defined" % name)
        solver = partial(DReal, "dreal/dreal4", LOGICS=DReal.LOGICS)
        solver.LOGICS = DReal.LOGICS
        env.factory._all_solvers[name] = solver
        env.factory._generic_solvers[name] = ("dreal/dreal4", DReal.LOGICS)
        env.factory.solver_preference_list.append(name)

        with Solver(name="dreal", logic=QF_NRA) as s:
            s.add_assertion(GT(Symbol("p", REAL), Symbol("q", REAL)))
            res1 = s.solve()
            s.push(1)
            s.add_assertion(GT(Symbol("r", REAL), Symbol("q", REAL)))
            res2 = s.solve()
            s.pop(1)
            s.add_assertion(Equals(Symbol("q", REAL), Real(1.0)))
            s.add_assertion(Equals(Symbol("p", REAL), Real(0.0)))
            res3 = s.solve()  # unsat
        assert res1 and res2 and not res3


if __name__ == "__main__":
    unittest.main()



import os
from pathlib import Path

from setuptools import find_packages, setup

with open(os.path.join("src", "funman_dreal", "_version.py")) as version_file:
    version = version_file.readlines()[-1].split()[-1].strip("\"'")

# read the contents of your README file
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="funman_dreal",
    version=version,
    description="Functional Model Analysis Tool - DReal Package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    author="Dan Bryce",
    author_email="dbryce@sift.net",
    license="MIT",
    packages=find_packages("src"),
    package_dir={"": "src"},
    install_requires=["funman", "docker", "tenacity", "pyparsing"],
    extras_require={"dreal": ["dreal"]},
    tests_require=["unittest"],
    zip_safe=False,
)



import functools

import dreal
from pysmt.decorators import catch_conversion_error
from pysmt.solvers.solver import (
    Converter,
    IncrementalTrackingSolver,
    Model,
    SolverOptions,
    UnsatCoreSolver,
)
from pysmt.walkers import DagWalker


class DRealConverter(Converter, DagWalker):
    def __init__(self, environment):
        DagWalker.__init__(self, environment)
        self.backconversion = {}
        self.mgr = environment.formula_manager
        self._get_type = environment.stc.get_type

        # Maps a Symbol into the corresponding internal yices instance
        self.symbol_to_decl = {}
        # Maps an internal yices instance into the corresponding symbol
        self.decl_to_symbol = {}

    @catch_conversion_error
    def convert(self, formula):
        return self.walk(formula)

    def walk_and(self, formula, args, **kwargs):
        res = functools.reduce(lambda a, b: dreal.And(a, b), args)
        # self._check_term_result(res)
        return res

    def walk_iff(self, formula, args, **kwargs):
        res = dreal.Iff(args[0], args[1])
        # self._check_term_result(res)
        return res

    def walk_or(self, formula, args, **kwargs):
        res = functools.reduce(lambda a, b: dreal.Or(a, b), args)
        # self._check_term_result(res)
        return res

    def walk_not(self, formula, args, **kwargs):
        res = dreal.Not(args[0])
        return res

    def walk_forall(self, formula, args, **kwargs):
        symbols = [self.walk(s) for s in formula._content.payload]
        res = dreal.forall(symbols, args[0])

        return res

    def walk_symbol(self, formula, **kwargs):
        symbol_type = formula.symbol_type()
        # var_type = self._type_to_dreal(symbol_type)
        if symbol_type.is_real_type():
            term = dreal.Variable(formula.symbol_name(), dreal.Variable.Real)
        elif symbol_type.is_bool_type():
            term = dreal.Variable(formula.symbol_name(), dreal.Variable.Bool)
        self.symbol_to_decl[formula] = term
        self.decl_to_symbol[term] = formula
        # yicespy.yices_set_term_name(term, formula.symbol_name())
        # self._check_term_result(term)
        return term

    def walk_le(self, formula, args, **kwargs):
        return args[0] <= args[1]

    def walk_lt(self, formula, args, **kwargs):
        return args[0] < args[1]

    def walk_plus(self, formula, args, **kwargs):
        return functools.reduce(lambda a, b: a + b, args)

    def walk_minus(self, formula, args, **kwargs):
        return args[0] - args[1]

    def walk_times(self, formula, args, **kwargs):
        res = functools.reduce(lambda a, b: a * b, args)
        # res = yicespy.yices_sum(len(args), args)
        # self._check_term_result(res)
        return res

    def walk_pow(self, formula, args, **kwargs):
        res = pow(args[0], args[1])
        return res

    def walk_equals(self, formula, args, **kwargs):
        res = args[0] == args[1]
        # tp = self._get_type(formula.arg(0))
        # res = None
        # if tp.is_bv_type():
        #     res = yicespy.yices_bveq_atom(args[0], args[1])
        # elif tp.is_int_type() or tp.is_real_type():
        #     res = yicespy.yices_arith_eq_atom(args[0], args[1])
        # else:
        #     assert tp.is_custom_type()
        #     res = yicespy.yices_eq(args[0], args[1])
        # self._check_term_result(res)
        return res

    def walk_bool_constant(self, formula, **kwargs):
        if formula.constant_value():
            return dreal.Formula.TRUE()
        else:
            return dreal.Formula.FALSE()

    def walk_real_constant(self, formula, **kwargs):
        frac = formula.constant_value()
        n, d = frac.numerator, frac.denominator
        res = float(n / d)
        # self._check_term_result(res)
        return res

    def _type_to_dreal(self, tp):
        if tp.is_bool_type():
            return dreal._dreal_py.Type.Bool
        elif tp.is_real_type():
            return dreal._dreal_py.Type.Real
        elif tp.is_int_type():
            return dreal._dreal_py.Type.Int
        elif tp.is_function_type():
            stps = [self._type_to_dreal(x) for x in tp.param_types]
            rtp = self._type_to_dreal(tp.return_type)
            raise NotImplementedError(tp)
            # arr = (yicespy.type_t * len(stps))(*stps)
            # return yicespy.yices_function_type(len(stps),
            #                                   stps,
            #                                   rtp)
        # elif tp.is_bv_type():
        #     return yicespy.yices_bv_type(tp.width)
        # elif tp.is_custom_type():
        #     return self.yicesSort(str(tp))
        else:
            raise NotImplementedError(tp)

    def get_dreal_ref(self, formula):
        # if formula.node_type in op.QUANTIFIERS:
        #     return z3.QuantifierRef
        # elif formula.node_type() in BOOLREF_SET:
        #     return z3.BoolRef
        # elif formula.node_type() in ARITHREF_SET:
        #     return z3.ArithRef
        # elif formula.node_type() in BITVECREF_SET:
        #     return z3.BitVecRef
        # el
        if formula.is_symbol() or formula.is_function_application():
            if formula.is_function_application():
                type_ = formula.function_name().symbol_type()
                type_ = type_.return_type
            else:
                type_ = formula.symbol_type()

            if type_.is_bool_type():
                return dreal._dreal_py.Type.Bool
            elif type_.is_real_type():
                return dreal._dreal_py.Type.Real
            elif type_.is_int_type():
                return dreal._dreal_py.Type.Int
            else:
                raise NotImplementedError(formula)
        elif formula.is_ite():
            child = formula.arg(1)
            return self.get_dreal_ref(child)
        else:
            assert formula.is_constant(), formula
            type_ = formula.constant_type()
            if type_.is_bool_type():
                return dreal._dreal_py.Type.Bool
            elif type_.is_real_type():
                return dreal._dreal_py.Type.Real
            else:
                raise NotImplementedError(formula)
